/***
 * Fractal ADL Parser
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 *
 * Contributors:
 * - Alessio Pace
 * - Philippe Merle (set the class loader to use during XML parsing)
 */

package org.objectweb.fractal.adl.xml;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.xml.sax.SAXException;

/**
 * Default implementation of the {@link XMLNodeFactory} interface.
 */
public class XMLNodeFactoryImpl implements XMLNodeFactory {
  /**
   * Cache of already read DTDs. Keys are DTD names, values are ASTClassLoader
   * objects.
   */
  protected Map<String, XMLNodeClassLoader> loaders = new HashMap<String, XMLNodeClassLoader>();

  /**
   * Class loader to use during XML parsing.
   */
  public ClassLoader classLoader = getClass().getClassLoader();

  // --------------------------------------------------------------------------
  // Implementation of the XMLNodeFactory interface
  // --------------------------------------------------------------------------

  public void checkDTD(final String systemId) throws SAXException {
    XMLNodeClassLoader loader = loaders.get(systemId);
    if (loader == null) {
      try {
        loader = new XMLNodeClassLoader(classLoader);

        InputStream is;
        // gets an InputStream to read the DTD
        if (systemId.startsWith("classpath://")) {
            is = classLoader.getResourceAsStream(
              systemId.substring("classpath://".length()));
          if (is == null)
            throw new SAXException("Unable to find DTD '" + systemId + "'");
        } else if (systemId.startsWith("file:")) {
          is = new URL(systemId).openStream();
        } else {
          throw new MalformedURLException("Unrecognized system identifier: "
              + systemId);
        }
        // verifies the DTD and initializes the AST class loader
        new DTDHandler().checkDTD(is, loader);
      } catch (final MalformedURLException e) {
        throw new SAXException("Cannot find the DTD", e);
      } catch (final IOException e) {
        throw new SAXException("Cannot read the DTD", e);
      } catch (final ClassNotFoundException e) {
        throw new SAXException("Cannot check the DTD", e);
      }

      loaders.put(systemId, loader);
    }
  }

  public XMLNode newXMLNode(final String systemId, final String qualifiedName)
      throws SAXException {
    XMLNodeClassLoader loader = loaders.get(systemId);
    if (loader == null) {
      checkDTD(systemId);
      loader = loaders.get(systemId);
    }
    final String xmlNodeClassName = loader.getASTClassName(qualifiedName);

    try {
      final Class loadClass = loader.loadClass(xmlNodeClassName);
      return (XMLNode) loadClass.newInstance();
    } catch (final ClassNotFoundException cnfe) {
      throw new SAXException(
          "The element <"
              + qualifiedName
              + "> can't be mapped to any AST node type. Check that you spell it correctly, or that the '"
              + xmlNodeClassName + "' exists, or that your DTD configuration ["
              + systemId + "] is correct", cnfe);
    } catch (final Exception e) {
      e.printStackTrace();
      throw new SAXException("Internal error", e);
    }
  }

  public ClassLoader getClassLoader() {
    return classLoader;
  }

  public void setClassLoader(final ClassLoader loader) {
    classLoader = loader;
  }
}
